#ifndef QUASSELC_H
#define QUASSELC_H

#include <glib.h>
#include "bot.h"
#include "cmds.h"
#include "display.h"
#include "endian.h"
#include "getters.h"
#include "io.h"
#include "setters.h"
#include "types.h"

#endif
